/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2021 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2021 Ryu Takegami (Twitter:@RyuTakegami)
 *
 *	[ CCPU ]
 *
 *	RHG
 *	  2002.03.11		I/OANZX̃fBČ悤ɂ
 *	  2002.05.06		FM-7[h̃u[gROM̈̈( ^^;)ύX
 *	  2003.11.20		XM7 V1.1Ή
 *	  2008.01.24		kanji.c̓{TuI/F؂藣ɊւύX
 *	  2010.06.18		FM-77[hɍ킹ăCmۗǂύX(V1̂)
 *	  2010.07.28		V2/V3FM77AV[hAV40/20ȍ~̃CjVG[^ROMg
 *						pۂɁAAV̓ƂĐVu[g̓]֎~
 *						ыu[g𗘗ppb`ZbgɓIɓĂ
 *						悤ɕύX (V340EX[hɂ͐Vu[g𗘗p܂)
 *						V2/V3̋@픻ʗ̈̐ݒ6oCgs悤ɕύX
 *	  2010.12.10		V3BOOT_MMR.ROMΉ
 *	  2012.04.24		FM-NEW7bg̐Vu[gROM(TL11-12)/FM-8u[g
 *						ROM(SM11-14)̍Npb`ɑΉ
 *	  2013.12.13		FM-NEW7bg̐Vu[gROM+F-BASIC V3.5Ƃgݍ
 *						킹(oהłł͂蓾Ȃ)ŃzbgZbgłȂ
 *						ۂɑΏ
 *	  2014.07.16		FM-7[ĥ݁ABASIC ROM̈ɏ񂾏ꍇɗRAM
 *						݂s悤ɎdlύX (Ȃ񂩂炵)
 *	  2021.04.18		AV40/40EX/40SX̐Vu[gɑ΂ẴhCu}bsO
 *						}RpCXCb`Őݒł悤ɂ
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "device.h"
#include "subctrl.h"
#include "ttlpalet.h"
#include "fdc.h"
#include "mainetc.h"
#include "multipag.h"
#include "kanji.h"
#include "tapelp.h"
#include "opn.h"
#include "mmr.h"
#include "apalet.h"
#include "rs232c.h"
#ifdef MIDI
#include "midi.h"
#endif
#if XM7_VER >= 2
#include "jcard.h"
#endif
#if XM7_VER >= 3
#include "dmac.h"
#endif
#if XM7_VER == 1
#ifdef JSUB
#include "jsubsys.h"
#endif
#ifdef BUBBLE
#include "bubble.h"
#endif
#endif
#include "ptm.h"
#include "mouse.h"

/*
 *	O[o [N
 */
BYTE *mainram_a;						/* RAM (\RAM)        $8000 */
BYTE *mainram_b;						/* RAM (RAM+)     $7C80 */
BYTE *basic_rom;						/* ROM (F-BASIC)      $7C00 */
BYTE *main_io;							/* CCPU I/O       $100 */
BOOL basicrom_en;						/* BASIC ROMCl[utO */
#if XM7_VER == 1
BYTE *basic_rom8;						/* ROM (F-BASIC V1.0) $7C00 */
BYTE *boot_bas;							/* u[g(BASIC,FM-7)  $200 */
BYTE *boot_dos;							/* u[g(DOS,FM-7)    $200 */
BYTE *boot_bas8;						/* u[g(BASIC,FM-8) $200 */
BYTE *boot_dos8;						/* u[g(DOS,FM-8)   $200 */
BYTE *boot_bbl8;						/* u[g(BBL,FM-8)   $200 */
#ifdef BUBBLE
BOOL bubble_available;					/* ougp\tO */
#endif
#endif
BYTE *boot_ram;							/* u[gRAM           $200 */
BOOL bootram_rw;						/* u[gRAM ݃tO */
BOOL rom_ram_write;						/* FM-7[hROM̈揑ύX */

BYTE *extram_a;							/* gRAM           $10000 */
#if XM7_VER >= 3
BYTE *extram_c;							/* AV40gRAM       $C0000 */
#endif

#if XM7_VER >= 2
BYTE *init_rom;							/* CjVG[^ROM    $2000 */
BOOL initrom_en;						/* CjVG[^ROMCl[utO */
#if XM7_VER >= 3
BOOL init_is_exsx;						/* CjVG[^ROM EX/SXtO */
#endif
#endif


/*
 *	X^eBbN [N
 */
#if XM7_VER >= 2
static BYTE *patch_branewboottfr;		/* Vu[g]pւBRA */
static BYTE *patch_jmpnewboot;			/* Vu[gւJMP */
#endif
static BYTE ioaccess_count;				/* I/ÖANZXJE^ */
static BOOL ioaccess_flag;				/* I/OANZXEFCgtO */

#if XM7_VER >= 3
/*
 *	Vu[gւ̃pb`1
 *	USART(̂̓pb`2),_hCuύX,IPLւ̃Wv
 */
static const BYTE bootpatch1[] = {
	0x0f, 0x10, 0xbd, 0x7f, 0x80, 0x86, 0x02, 0x97,
	0x10, 0x4f, 0xbc, 0xfb, 0xfe, 0x27, 0x19, 0xe6,
	0xe4, 0xc1, 0xff, 0x27, 0x13, 0xc1, 0x01, 0x23,
	0x0f, 0xcc, 0x00, 0xfd, 0x1f, 0x9b, 0x17, 0xfe,
	0x2b, 0x1f, 0x8b, 0x4c, 0xe6, 0xe4, 0xc0, 0x02,
	0x1f, 0x8b, 0x6e, 0x84,
};

/*
 *	Vu[gւ̃pb`2
 *	USART(pb`1̑)
 */
static const BYTE bootpatch2[] = {
	0xce, 0xfd, 0x06, 0x8d, 0x0d, 0xce, 0xfd, 0x24,
	0x86, 0x04, 0x8d, 0x06, 0x33, 0x42, 0x4a, 0x26,
	0xf9, 0x39, 0x6f, 0x41, 0x6f, 0x41, 0x6f, 0x41,
	0xc6, 0x40, 0xe7, 0x41, 0x39,
};
#endif


/*
 *	CCPU
 *	
 */
BOOL FASTCALL mainmem_init(void)
{
	int i;
	BYTE *p;
	BOOL hidden_bootrom;
#if XM7_VER == 1
	int addr;
#endif

	/* xASăNA */
	mainram_a = NULL;
	mainram_b = NULL;
	basic_rom = NULL;
	main_io = NULL;
	extram_a = NULL;
#if XM7_VER == 1
	boot_bas = NULL;
	boot_dos = NULL;
	boot_bas8 = NULL;
	boot_dos8 = NULL;
	boot_bbl8 = NULL;
#ifdef BUBBLE
	bubble_available = TRUE;
#endif
#else
	patch_branewboottfr = NULL;
	patch_jmpnewboot = NULL;
#if XM7_VER >= 3
	extram_c = NULL;
#endif
	init_rom = NULL;
#endif
	boot_ram = NULL;
	rom_ram_write = TRUE;

	/* RAM */
	mainram_a = (BYTE *)malloc(0x8000);
	if (mainram_a == NULL) {
		return FALSE;
	}
#if XM7_VER == 1
	mainram_b = (BYTE *)malloc(0x7e00);
#else
	mainram_b = (BYTE *)malloc(0x7c80);
#endif
	if (mainram_b == NULL) {
		return FALSE;
	}

	/* BASIC ROM, I/O */
	basic_rom = (BYTE *)malloc(0x8000);
	if (basic_rom == NULL) {
		return FALSE;
	}
#if XM7_VER == 1
	basic_rom8 = (BYTE *)malloc(0x7c00);
	if (basic_rom8 == NULL) {
		return FALSE;
	}
#endif
	main_io = (BYTE *)malloc(0x0100);
	if (main_io == NULL) {
		return FALSE;
	}

	/* gRAMACjVG[^ROM */
#if XM7_VER >= 2
	extram_a = (BYTE *)malloc(0x10000);
	if (extram_a == NULL) {
		return FALSE;
	}
#if XM7_VER >= 3
	extram_c = (BYTE *)malloc(0xc0000);
	if (extram_c == NULL) {
		return FALSE;
	}
#endif
	init_rom = (BYTE *)malloc(0x2000);
	if (init_rom == NULL) {
		return FALSE;
	}
#else
	/* V1 (400CZbgC192KB) */
	extram_a = (BYTE *)malloc(0x30000);
	if (extram_a == NULL) {
		return FALSE;
	}
#endif

	/* u[gROM/RAM */
#if XM7_VER == 1
	boot_bas = (BYTE *)malloc(0x200);
	if (boot_bas == NULL) {
		return FALSE;
	}
	boot_dos = (BYTE *)malloc(0x200);
	if (boot_dos == NULL) {
		return FALSE;
	}
	boot_bas8 = (BYTE *)malloc(0x200);
	if (boot_bas8 == NULL) {
		return FALSE;
	}
	boot_dos8 = (BYTE *)malloc(0x200);
	if (boot_dos8 == NULL) {
		return FALSE;
	}
	boot_bbl8 = (BYTE *)malloc(0x200);
	if (boot_bbl8 == NULL) {
		return FALSE;
	}
#endif
	boot_ram = (BYTE *)malloc(0x200);
	if (boot_ram == NULL) {
		return FALSE;
	}

	/* ROMt@Cǂݍ */
	hidden_bootrom = TRUE;
	if (!file_load(FBASIC_ROM, basic_rom, 0x8000)) {
		hidden_bootrom = FALSE;
		if (!file_load(FBASIC_ROM, basic_rom, 0x7c00)) {
#if XM7_VER == 1
			available_fm7roms = FALSE;
#else
			return FALSE;
#endif
		}
	}
#if XM7_VER == 1
	if (!file_load(FBASIC10_ROM, basic_rom8, 0x7c00)) {
		available_fm8roms = FALSE;
	}
#endif

#if XM7_VER >= 2
	if (!file_load(INITIATE_ROM, init_rom, 0x2000)) {
		return FALSE;
	}

	/* u[gN(AVV[Y) */
	for (i=0; i<2; i++) {
		p = &init_rom[0x1800 + i * 0x200];
		if (p[0x14f] == 0x26) {
			p[0x14f] = 0x21;
		}
		if (p[0x153] == 0x26) {
			p[0x153] = 0x21;
		}
	}

#if XM7_VER >= 3
	/* Vu[gNEhCuΉύX}(AV20EX/AV40EX/AV40SX) */
	p = &init_rom[0x1c00];
#ifndef DRIVESWAP
	if ((p[0x8f] == 0x10) && (p[0x90] == 0x23)) {
		/* 擪2oCgĂȂǁA܂c */
		memcpy(&p[0x0247], bootpatch1, sizeof(bootpatch1));
		memcpy(&p[0x0380], bootpatch2, sizeof(bootpatch2));
		p[0x08f] = 0x16;
		p[0x090] = 0x01;
		p[0x091] = 0xb5;
		if (p[0x166] == 0x27) {
			p[0x166] = 0x21;
		}
		if (p[0x1d6] == 0x26) {
			p[0x1d6] = 0x21;
		}
	}

	/* Vu[gNEhCuΉύX}(AV40/AV20) */
	if ((p[0x74] == 0x10) && (p[0x75] == 0x23)) {
		/* 擪2oCgĂȂǁA܂c */
		memcpy(&p[0x0207], bootpatch1, sizeof(bootpatch1));
		memcpy(&p[0x0380], bootpatch2, sizeof(bootpatch2));
		p[0x074] = 0x16;
		p[0x075] = 0x01;
		p[0x076] = 0x90;
		p[0x227] = 0x50;	/* pb`AV20EX/40EX/40SXƂ̈Ⴂ͂ꂾ */
		if (p[0xf3] == 0x27) {
			p[0xf3] = 0x21;
		}
		if (p[0x196] == 0x26) {
			p[0x196] = 0x21;
		}
	}
#endif
#endif

	/* Vu[g֘ÃAhXEL */
	for (i=0; i<0xb00; i++) {
		/* Vu[g]ւBRA */
		if (!patch_branewboottfr) {
			if ((init_rom[i + 0] == 0x20) && (init_rom[i + 1] == 0xd7)) {
				patch_branewboottfr = &init_rom[i];
			}
		}

		/* Vu[gւJMP(CjVG[^̖ɑ݂) */
		if (!patch_jmpnewboot) {
			if ((init_rom[i + 0] == 0x7e) && (init_rom[i + 1] == 0x50) &&
				(init_rom[i + 2] == 0x00)) {
				patch_jmpnewboot = &init_rom[i];
			}
		}

		/* 烋[v𔲂 */
		if (patch_branewboottfr && patch_jmpnewboot) {
			break;
		}
	}
#if XM7_VER >= 2
	if (!hidden_bootrom) {
		if (!file_load(BOOTHIDDEN_ROM, &basic_rom[0x7e00], 0x200)) {
			memcpy(&basic_rom[0x7e00], &init_rom[0x1800], 0x1e0);
			memset(&basic_rom[0x7fe0], 0, 30);
			basic_rom[0x7ffe] = 0xfe;
			basic_rom[0x7fff] = 0x00;
		}
	}
#endif
#else
	if (!file_load(BOOTBAS_ROM, boot_bas, 0x1e0)) {
		available_fm7roms = FALSE;
	}
	if (!file_load(BOOTDOS_ROM, boot_dos, 0x1e0)) {
		available_fm7roms = FALSE;
	}
	if (!hidden_bootrom) {
		if (!file_load(BOOTHIDDEN_ROM, &basic_rom[0x7e00], 0x200)) {
			memcpy(&basic_rom[0x7e00], boot_bas, 0x1e0);
			memset(&basic_rom[0x7fe0], 0, 30);
			basic_rom[0x7ffe] = 0xfe;
			basic_rom[0x7fff] = 0x00;
		}
	}
	if (!file_load(BOOTBAS8_ROM, boot_bas8, 0x1e0)) {
		available_fm8roms = FALSE;
	}
	if (!file_load(BOOTDOS8_ROM, boot_dos8, 0x1e0)) {
		available_fm8roms = FALSE;
	}
#ifdef BUBBLE
	if (!file_load(BOOTBBL8_ROM, boot_bbl8, 0x1e0)) {
		bubble_available = FALSE;
	}
#endif

	/* u[gN(FM-7/NEW7(1st lot)/8) */
	for (i=0; i<5; i++) {
		switch (i) {
			case 0:		p = boot_bas;
						break;
			case 1:		p = boot_dos;
						break;
			case 2:		p = boot_bas8;
						break;
			case 3:		p = boot_dos8;
						break;
			case 4:		p = boot_bbl8;
						break;
			default:	ASSERT(FALSE);
						break;
		}

		/* BOOT ROMEFCg[`ĎKp */
		for (addr = 0x0000; addr < 0x01dd; addr++) {
			if ((p[addr + 0] == 0x31) && (p[addr + 1] == 0x3f) &&
				(p[addr + 2] == 0x26) && (p[addr + 3] == 0xfc)) {
				p[addr + 2] = 0x21;
			}
		}
		p[0x1fe] = 0xfe;
		p[0x1ff] = 0x00;
	}
#endif

	/* u[gRAMNA */
	/* (_]΍EZbgɑS̈ȂȂ) */
	memset(boot_ram, 0, 0x200);

#if XM7_VER >= 3
	/* EX/SX̃CjVG[^ROM`FbN */
	if (init_rom[0xb10] == '1') {
		init_is_exsx = TRUE;
	}
	else {
		init_is_exsx = FALSE;
	}
#endif

#if XM7_VER == 1
	/* FM-7/FM-8ǂROMsSȏꍇAsƂ */
	if (!available_fm8roms && !available_fm7roms) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	CCPU
 *	N[Abv
 */
void FASTCALL mainmem_cleanup(void)
{
/*
	ASSERT(mainram_a);
	ASSERT(mainram_b);
	ASSERT(basic_rom);
	ASSERT(main_io);
	ASSERT(extram_a);
#if XM7_VER >= 3
	ASSERT(extram_c);
#endif
#if XM7_VER >= 2
	ASSERT(init_rom);
#endif

#if XM7_VER == 1
	ASSERT(basic_rom8);
	ASSERT(boot_bas);
	ASSERT(boot_dos);
	ASSERT(boot_bas8);
	ASSERT(boot_dos8);
	ASSERT(boot_bbl8);
#endif
	ASSERT(boot_ram);
*/

	/* rŎsꍇl */
	if (mainram_a) {
		free(mainram_a);
	}
	if (mainram_b) {
		free(mainram_b);
	}
	if (basic_rom) {
		free(basic_rom);
	}
	if (main_io) {
		free(main_io);
	}

	if (extram_a) {
		free(extram_a);
	}
#if XM7_VER >= 3
	if (extram_c) {
		free(extram_c);
	}
#endif
#if XM7_VER >= 2
	if (init_rom) {
		free(init_rom);
	}
#endif

#if XM7_VER == 1
	if (basic_rom8) {
		free(basic_rom8);
	}
	if (boot_bas) {
		free(boot_bas);
	}
	if (boot_dos) {
		free(boot_dos);
	}
	if (boot_bas8) {
		free(boot_bas8);
	}
	if (boot_dos8) {
		free(boot_dos8);
	}
	if (boot_bbl8) {
		free(boot_bbl8);
	}
#endif
	if (boot_ram) {
		free(boot_ram);
	}
}

/*
 *	CCPU
 *	Zbg
 */
void FASTCALL mainmem_reset(void)
{
	/* I/OԁEI/OANZXJE^ */
	memset(main_io, 0xff, 0x0100);
	ioaccess_count = 0;
	ioaccess_flag = FALSE;

	/* BASIC[hł΁AF-BASIC ROMCl[u */
	if (boot_mode == BOOT_BASIC) {
		basicrom_en = TRUE;
	}
	else {
		basicrom_en = FALSE;
	}

#if XM7_VER >= 2
	if (fm7_ver >= 2) {
		/* AV/40EX[h : CjVG[^ONAu[gRAM݉ */
		initrom_en = TRUE;
		bootram_rw = TRUE;
	}
	else {
		/* FM-7[h : CjVG[^OFFAu[gRAM݋֎~ */
		initrom_en = FALSE;
		bootram_rw = FALSE;
	}

	/* u[gRAM ZbgAbv */
	/* 荞݃xN^ȊOUׂăNA(_]΍) */
	memset(boot_ram, 0, 0x1f0);

	/* FM-7[hɂ̓u[gRAM̓e] */
	mainmem_transfer_boot();

	/* CjVG[^ROM n[hEFAo[W */
	switch (fm7_ver) {
		case 1:
			/* FM-7 */
			break;
		case 2:
			/* FM77AV */
			memset(&init_rom[0x0b0e], 0xff, 6);

			/* CjVG[^FM77AV̓ƂȂ悤Ƀpb` */
			if (patch_branewboottfr && patch_jmpnewboot) {
				patch_branewboottfr[0x0000] = 0x21;
				patch_jmpnewboot[0x0001] = 0xfe;
				patch_jmpnewboot[0x0002] = 0x00;
			}
			break;
#if XM7_VER >= 3
		case 3:
			/* FM77AV40EX */
			init_rom[0xb0e] = '4';
			init_rom[0xb0f] = '0';
			init_rom[0xb10] = '1';
			init_rom[0xb11] = 'M';
			init_rom[0xb12] = 'a';
			init_rom[0xb13] = '.';

			/* CjVG[^FM77AV40EX{̓ƂȂ悤ɖ߂ */
			if (patch_branewboottfr && patch_jmpnewboot) {
				patch_branewboottfr[0x0000] = 0x20;
				patch_jmpnewboot[0x0001] = 0x50;
				patch_jmpnewboot[0x0002] = 0x00;
			}
			break;
#endif
	}
#else
	/* u[g̈揑݋֎~ */
	bootram_rw = FALSE;
#endif
}

/*
 *	FM-7[hp u[gRAM]
 */
#if XM7_VER >= 2
void FASTCALL mainmem_transfer_boot(void)
{
	if (fm7_ver == 1) {
		if (boot_mode == BOOT_BASIC) {
			memcpy(boot_ram, &init_rom[0x1800], 0x1e0);
		}
		else {
			memcpy(boot_ram, &init_rom[0x1a00], 0x1e0);
		}
		boot_ram[0x1fe] = 0xfe;
		boot_ram[0x1ff] = 0x00;
	}
}
#endif

/*
 *	I/OANZX̃fB
 */
static void FASTCALL mainmem_iowait(void)
{
	BYTE tmp;

#if XM7_VER == 1
	/* 1.2MHz[hłI/OEFCgȂ */
	if (lowspeed_mode) {
		return;
	}
#endif

	/* NbN΂ɕKvI/OANZX񐔂ݒ */
#if XM7_VER >= 3
	if ((mmr_flag || twr_flag) && !mmr_fastmode && ioaccess_flag) {
#else
	if ((mmr_flag || twr_flag) && ioaccess_flag) {
#endif
		/* MMRI(2)  3̃ANZX1TCN */
		tmp = 3;
	}
	else {
		/* MMRIt/[h/MMRI(1)  2̃ANZX1TCN */
		tmp = 2;
	}

	/* K񐔈ȏI/OANZX΁ANbN΂ */
	ioaccess_count ++;
	if (ioaccess_count >= tmp) {
		maincpu.total ++;
		ioaccess_count = 0;

		ioaccess_flag = !ioaccess_flag;
	}
}


/*
 *	CCPU
 *	PoCg擾
 */
BYTE FASTCALL mainmem_readb(WORD addr)
{
	BYTE dat;

	/* MMR, TWR`FbN */
	if (mmr_flag || twr_flag) {
		/* MMRATWRʂ */
		if (mmr_extrb(&addr, &dat)) {
			return dat;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		return mainram_a[addr];
	}
	if (addr < 0x8000) {
		if (initrom_en) {
			return init_rom[addr - 0x6000];
		}
		else {
			return mainram_a[addr];
		}
	}
#else
	if (addr < 0x8000) {
		return mainram_a[addr];
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return basic_rom8[addr - 0x8000];
			}
#endif
			return basic_rom[addr - 0x8000];
		}
		else {
			return mainram_b[addr - 0x8000];
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		return mainram_b[addr - 0x8000];
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			return shared_ram[(WORD)(addr - 0xfc80)];
		}
		else {
			return 0xff;
		}
	}

	/* u[gROM/RAM */
#if XM7_VER >= 2
	if ((addr >= 0xfffe) && (initrom_en)) {
		/* ZbgxN^ */
		mainmem_iowait();
		return init_rom[addr - 0xe000];
	}
#endif
	if (addr >= 0xfe00) {
		if (addr <= 0xffdf) {
			mainmem_iowait();
		}
#if XM7_VER >= 2
		return boot_ram[addr - 0xfe00];
#else
		if (((addr >= 0xffe0) && (addr < 0xfffe)) ||
			((fm_subtype == FMSUB_FM77) && bootram_rw)) {
			/* FM-77 RAM */
			return boot_ram[addr - 0xfe00];
		}
		else {
			if (fm_subtype == FMSUB_FM8) {
				/* FM-8 BOOT ROM */
				if (boot_mode == BOOT_BASIC) {
					return boot_bas8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS) {
					return boot_dos8[addr - 0xfe00];
				}
				else {
					return boot_bbl8[addr - 0xfe00];
				}
			}
			else {
				/* FM-7 BOOT ROM */
				if (boot_mode == BOOT_BASIC) {
					return boot_bas[addr - 0xfe00];
				}
				else {
					return boot_dos[addr - 0xfe00];
				}
			}
		}
#endif
	}

	/*
	 *	I/O
	 */
	mainmem_iowait();
	if (mainetc_readb(addr, &dat)) {
		return dat;
	}
	if (ttlpalet_readb(addr, &dat)) {
		return dat;
	}
	if (subctrl_readb(addr, &dat)) {
		return dat;
	}
	if (multipag_readb(addr, &dat)) {
		return dat;
	}
	if (fdc_readb(addr, &dat)) {
		return dat;
	}
	if (kanji_readb(addr, &dat)) {
		return dat;
	}
	if (tapelp_readb(addr, &dat)) {
		return dat;
	}
	if (opn_readb(addr, &dat)) {
		return dat;
	}
	if (whg_readb(addr, &dat)) {
		return dat;
	}
	if (thg_readb(addr, &dat)) {
		return dat;
	}
	if (mmr_readb(addr, &dat)) {
		return dat;
	}
#if XM7_VER >= 2
	if (apalet_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef MIDI
	if (midi_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef RSC
	if (rs232c_readb(addr, &dat)) {
		return dat;
	}
#endif
#if XM7_VER >= 3
	if (dmac_readb(addr, &dat)) {
		return dat;
	}
#endif
#if XM7_VER == 1
#ifdef JSUB
	if (jsub_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef BUBBLE
	if (bmc_readb(addr, &dat)) {
		return dat;
	}
#endif
#endif
#ifdef MOUSE
	if (ptm_readb(addr, &dat)) {
		return dat;
	}
	if (mouse_readb(addr, &dat)) {
		return dat;
	}
#endif

	return 0xff;
}

/*
 *	CCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL mainmem_readbnio(WORD addr)
{
	BYTE dat;

	/* MMR, TWR`FbN */
	if (mmr_flag || twr_flag) {
		/* MMRATWRʂ */
		if (mmr_extbnio(&addr, &dat)) {
			return dat;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		return mainram_a[addr];
	}
	if (addr < 0x8000) {
		if (initrom_en) {
			return init_rom[addr - 0x6000];
		}
		else {
			return mainram_a[addr];
		}
	}
#else
	if (addr < 0x8000) {
		return mainram_a[addr];
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return basic_rom8[addr - 0x8000];
			}
#endif
			return basic_rom[addr - 0x8000];
		}
		else {
			return mainram_b[addr - 0x8000];
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		return mainram_b[addr - 0x8000];
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			return shared_ram[(WORD)(addr - 0xfc80)];
		}
		else {
			return 0xff;
		}
	}

	/* u[gROM/RAM */
#if XM7_VER >= 2
	if ((addr >= 0xfffe) && (initrom_en)) {
		/* ZbgxN^ */
		return init_rom[addr - 0xe000];
	}
#endif
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		return boot_ram[addr - 0xfe00];
#else
		if (((addr >= 0xffe0) && (addr < 0xfffe)) ||
			((fm_subtype == FMSUB_FM77) && bootram_rw)) {
			return boot_ram[addr - 0xfe00];
		}
		else {
			if (fm_subtype == FMSUB_FM8) {
				if (boot_mode == BOOT_BASIC) {
					return boot_bas8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS) {
					return boot_dos8[addr - 0xfe00];
				}
				else {
					return boot_bbl8[addr - 0xfe00];
				}
			}
			else {
				if (boot_mode == BOOT_BASIC) {
					return boot_bas[addr - 0xfe00];
				}
				else {
					return boot_dos[addr - 0xfe00];
				}
			}
		}
#endif
	}

	/* I/O */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	return main_io[addr - 0xfd00];
}

/*
 *	CCPU
 *	PoCg
 */
void FASTCALL mainmem_writeb(WORD addr, BYTE dat)
{
	/* MMR, TWR`FbN */
	if (mmr_flag || twr_flag) {
		/* MMRATWRʂ */
		if (mmr_extwb(&addr, dat)) {
			return;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		mainram_a[addr] = dat;
		return;
	}
	if (addr < 0x8000) {
		if (!initrom_en) {
			mainram_a[addr] = dat;
		}
		return;
	}
#else
	if (addr < 0x8000) {
		mainram_a[addr] = dat;
		return;
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
			/* ROMRCBBIOSĂяoP[X */
#if XM7_VER == 1
			if ((fm_subtype != FMSUB_FM8) && rom_ram_write) {
#else
			if ((fm7_ver == 1) && rom_ram_write) {
#endif
				/* FM-7łF-BASIC ROMIɂRAMɏ߂悤Ȃ̂ */
				mainram_b[addr - 0x8000] = dat;
			}
			return;
		}
		else {
			mainram_b[addr - 0x8000] = dat;
			return;
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		mainram_b[addr - 0x8000] = dat;
		return;
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			shared_ram[(WORD)(addr - 0xfc80)] = dat;
			return;
		}
		else {
			/* BASE09΍ */
			return;
		}
	}

	/* u[gRAM */
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		if ((bootram_rw) && (fm7_ver >= 2)) {
#else
		if ((fm_subtype == FMSUB_FM77) && bootram_rw) {
#endif
			boot_ram[addr - 0xfe00] = dat;
			if ((addr <= 0xffdf) || (addr >= 0xfffe)) {
				mainmem_iowait();
			}
			return;
		}

		/* u[g[NRAMAxN^ */
		if ((addr >= 0xffe0) && (addr < 0xfffe)) {
			boot_ram[addr - 0xfe00] = dat;
		}
		return;
	}

	/*
	 *	I/O
	 */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	main_io[(WORD)(addr - 0xfd00)] = dat;
	mainmem_iowait();

	if (mainetc_writeb(addr, dat)) {
		return;
	}
	if (ttlpalet_writeb(addr, dat)) {
		return;
	}
	if (subctrl_writeb(addr, dat)) {
		return;
	}
	if (multipag_writeb(addr, dat)) {
		return;
	}
	if (fdc_writeb(addr, dat)) {
		return;
	}
	if (kanji_writeb(addr, dat)) {
		return;
	}
	if (tapelp_writeb(addr, dat)) {
		return;
	}
	if (opn_writeb(addr, dat)) {
		return;
	}
	if (whg_writeb(addr, dat)) {
		return;
	}
	if (thg_writeb(addr, dat)) {
		return;
	}
	if (mmr_writeb(addr, dat)) {
		return;
	}
#if XM7_VER >= 2
	if (apalet_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef MIDI
	if (midi_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef RSC
	if (rs232c_writeb(addr, dat)) {
		return;
	}
#endif
#if XM7_VER >= 3
	if (dmac_writeb(addr, dat)) {
		return;
	}
#endif
#if XM7_VER == 1
#ifdef JSUB
	if (jsub_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef BUBBLE
	if (bmc_writeb(addr, dat)) {
		return;
	}
#endif
#endif
#ifdef MOUSE
	if (ptm_writeb(addr, dat)) {
		return;
	}
	if (mouse_writeb(addr, dat)) {
		return;
	}
#endif

	return;
}

/*
 *	CCPUAhXoX
 *	PoCgǂݏo(I/OȂ)
 */
BYTE FASTCALL mainmem_readbnio_p(DWORD raddr)
{
	DWORD rsegment;
	WORD addr;

	rsegment = (raddr & 0xf0000);
	addr = (WORD)(raddr & 0xffff);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER == 1
		/* BASIC ROMRAMELRAMEI/ÖɂROM/RAM */
		if ((raddr >= 0x3fc00) && (raddr < 0x3fe00)) {
			if (basicrom_en) {
				return basic_rom[raddr - 0x38000];
			}
			return mainram_b[raddr - 0x38000];
		}

		/* u[g̈ɂROM/RAM */
		if (raddr >= 0x3fe00) {
			if (basicrom_en) {
				return basic_rom[raddr - 0x38000];
			}
			return boot_ram[raddr - 0x3fe00];
		}
#endif

		/* $30ZOg */
		/* CRAM(\) */
#if XM7_VER >= 2
		if (addr < 0x6000) {
			return mainram_a[addr];
		}
		if (addr < 0x8000) {
			if (initrom_en) {
				return init_rom[addr - 0x6000];
			}
			return mainram_a[addr];
		}
#else
		if (addr < 0x8000) {
			return mainram_a[addr];
		}
#endif

		/* BASIC ROM or CRAM() */
		if (addr < 0xfc00) {
			if (basicrom_en) {
#if XM7_VER == 1
				if (fm_subtype == FMSUB_FM8) {
					return basic_rom8[addr - 0x8000];
				}
#endif
				return basic_rom[addr - 0x8000];
			}
			return mainram_b[addr - 0x8000];
		}

		/* 풓̈̏ */
		return mainmem_readbnio(addr);
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		return extram_a[raddr & 0xffff];
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			return submem_readbnio((WORD)(raddr & 0xffff));
		}
		return 0xff;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
#if XM7_VER >= 2
		return jcard_readb((WORD)(raddr & 0xffff));
#else
		return 0xff;
#endif
	}

#if XM7_VER >= 3
	/* 768KB gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram) {
			return extram_c[raddr - 0x40000];
		}
		return 0xff;
	}
#endif

	/* ܂Ȃ */
	ASSERT(FALSE);
	return 0xff;
#elif XM7_VER == 1
	/* FM-77 gRAM */
	if (mmr_64kbmode) {
		if ((rsegment >> 16) == (DWORD)bnk_reg) {
			return extram_a[(DWORD)((raddr & 0xffff) | 0x20000)];
		}
		return 0xff;
	}

	return extram_a[raddr];
#endif
}

/*
 *	CCPUAhXoX
 *	PoCg
 */
void FASTCALL mainmem_writeb_p(DWORD raddr, BYTE dat)
{
	DWORD rsegment;
	WORD addr;

	rsegment = (raddr & 0xf0000);
	addr = (WORD)(raddr & 0xffff);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER == 1
		/* BASIC ROMRAMELRAMEI/ÖɂROM/RAM */
		if ((raddr >= 0x3fc00) && (raddr < 0x3fe00)) {
			if (!basicrom_en) {
				mainram_b[raddr - 0x38000] = dat;
			}
			return;
		}

		/* u[g̈ɂROM/RAM */
		if (raddr >= 0x3fe00) {
			if (!basicrom_en) {
				boot_ram[raddr - 0x3fe00] = dat;
			}
			return;
		}
#endif

		/* $30ZOg */
		/* CRAM(\) */
#if XM7_VER >= 2
		if (addr < 0x6000) {
			mainram_a[addr] = dat;
			return;
		}
		if (addr < 0x8000) {
			if (!initrom_en) {
				mainram_a[addr] = dat;
			}
			return;
		}
#else
		if (addr < 0x8000) {
			mainram_a[addr] = dat;
			return;
		}
#endif

		/* BASIC ROM or CRAM() */
		if (addr < 0xfc00) {
			if (basicrom_en) {
				/* ROMRCBBIOSĂяoP[X */
#if XM7_VER == 1
			if ((fm_subtype != FMSUB_FM8) && rom_ram_write) {
#else
			if ((fm7_ver == 1) && rom_ram_write) {
#endif
					/* FM-7łBASIC ROMIɂRAMɏ߂悤Ȃ̂ */
					mainram_b[addr - 0x8000] = dat;
				}
				return;
			}
			else {
				mainram_b[addr - 0x8000] = dat;
				return;
			}
		}

		/* 풓̈̏ */
		mainmem_writeb(addr, dat);
		return;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		extram_a[raddr & 0xffff] = dat;
		return;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			submem_writeb((WORD)(raddr & 0xffff), dat);
		}
		return;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
#if XM7_VER >= 2
		jcard_writeb((WORD)(raddr & 0xffff), dat);
#endif
		return;
	}

#if XM7_VER >= 3
	/* AV40gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram) {
			extram_c[raddr - 0x40000] = dat;
		}
		return;
	}
#endif
#else
	/* FM-77 gRAM */
	if (mmr_64kbmode) {
		if ((rsegment >> 16) == (DWORD)bnk_reg) {
			extram_a[(DWORD)((raddr & 0xffff) | 0x20000)] = dat;
		}
		return;
	}

	extram_a[raddr] = dat;
	return;
#endif
}

/*
 *	CCPU
 *	Z[u
 */
BOOL FASTCALL mainmem_save(int fileh)
{
	if (!file_write(fileh, mainram_a, 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_write(fileh, mainram_b, 0x7e00)) {
#else
	if (!file_write(fileh, mainram_b, 0x7c80)) {
#endif
		return FALSE;
	}

	if (!file_write(fileh, main_io, 0x100)) {
		return FALSE;
	}

	if (!file_write(fileh, extram_a, 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x8000], 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_write(fileh, &extram_a[0x10000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x18000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x20000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x28000], 0x8000)) {
		return FALSE;
	}
#endif
	if (!file_write(fileh, boot_ram, 0x200)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, basicrom_en)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_bool_write(fileh, initrom_en)) {
		return FALSE;
	}
#endif
	if (!file_bool_write(fileh, bootram_rw)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if (!file_write(fileh, &init_rom[0x0b0e], 3)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	CCPU
 *	[h
 */
BOOL FASTCALL mainmem_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	if (!file_read(fileh, mainram_a, 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	/* Ver302g */
	if (ver <= 301) {
		if (!file_read(fileh, mainram_b, 0x7c80)) {
			return FALSE;
		}
		memset(&mainram_b[0x7c80], 0, 0x180);
	}
	else {
		if (!file_read(fileh, mainram_b, 0x7e00)) {
			return FALSE;
		}
	}
#else
	if (!file_read(fileh, mainram_b, 0x7c80)) {
		return FALSE;
	}
#endif

	if (!file_read(fileh, main_io, 0x100)) {
		return FALSE;
	}

	if (!file_read(fileh, extram_a, 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x8000], 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_read(fileh, &extram_a[0x10000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x18000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x20000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x28000], 0x8000)) {
		return FALSE;
	}
#endif
	if (!file_read(fileh, boot_ram, 0x200)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &basicrom_en)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_bool_read(fileh, &initrom_en)) {
		return FALSE;
	}
#endif
	if (!file_bool_read(fileh, &bootram_rw)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if (ver >= 800) {
		if (!file_read(fileh, &init_rom[0x0b0e], 3)) {
			return FALSE;
		}
	}
#endif

	return TRUE;
}
